
/* Chrysalide - Outil d'analyse de fichiers binaires
 * asm.c - implémentations génériques de fonctionnalités spécifiques
 *
 * Copyright (C) 2014-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "asm.h"



/******************************************************************************
*                                                                             *
*  Paramètres  : v = valeur quelconque sur 32 bits.                           *
*                p = position du premier bit à 1 (poids fort). [OUT]          *
*                                                                             *
*  Description : Détermine l'indice du premier bit à 1, côté gauche.          *
*                                                                             *
*  Retour      : true si le nombre est différent de zéro, false sinon.        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool msb_32(uint32_t v, unsigned int *p)
{
    static const unsigned int bval[] = { 0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4 };

    /* S'il n'y a aucun bit à 1... */
    if (v == 0) return false;

    /**
     * Il existe de nombreuses méthodes pour obtenir le résultat attendu
     * sans recourir à des extensions GCC ou à des instructions d'assembleur :
     *
     *  - http://stackoverflow.com/questions/2589096/find-most-significant-bit-left-most-that-is-set-in-a-bit-array
     *  - http://graphics.stanford.edu/~seander/bithacks.html#IntegerLogObvious
     *
     */

    *p = 0;

    if (v & 0xffff0000) { *p += 16 / 1; v >>= 16 / 1; }
    if (v & 0x0000ff00) { *p += 16 / 2; v >>= 16 / 2; }
    if (v & 0x000000f0) { *p += 16 / 4; v >>= 16 / 4; }

    *p += bval[v];

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : v = valeur quelconque sur 64 bits.                           *
*                p = position du premier bit à 1 (poids fort). [OUT]          *
*                                                                             *
*  Description : Détermine l'indice du premier bit à 1, côté gauche.          *
*                                                                             *
*  Retour      : true si le nombre est différent de zéro, false sinon.        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool msb_64(uint64_t v, unsigned int *p)
{
    static const unsigned int bval[] = { 0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4 };

    /* S'il n'y a aucun bit à 1... */
    if (v == 0) return false;

    /**
     * Cf. msb_32().
     */

    *p = 0;

    if (v & 0xffffffff00000000ull) { *p += 32 / 1; v >>= 32 / 1; }
    if (v & 0x00000000ffff0000ull) { *p += 32 / 2; v >>= 32 / 2; }
    if (v & 0x000000000000ff00ull) { *p += 32 / 4; v >>= 32 / 4; }
    if (v & 0x00000000000000f0ull) { *p += 32 / 8; v >>= 32 / 8; }

    *p += bval[v];

    return true;

}
